@@ -3,9 +3,9 @@
 -- Dungeoneering functions.
 ------------------------------------------------------------------------------
 
-require("dlua/point.lua")
-require("dlua/fnwrap.lua")
-require('dlua/util.lua')
+crawl_require("dlua/point.lua")
+crawl_require("dlua/fnwrap.lua")
+crawl_require('dlua/util.lua')
 
 -- Namespace for callbacks (just an aid to recognising callbacks, no magic)
 util.namespace('callback')
@@ -142,7 +142,7 @@ function dgn_map_copy_hooks_from(hook_source_mapname, hook_name)
     return
   end
 
-  local target_environment = dgn.map_environment();
+  local target_environment = dgn.map_environment()
   local target_hook = dgn_fixup_hook(target_environment[hook_variable])
   if not target_hook then
     target_hook = source_hook
@@ -250,7 +250,7 @@ function dgn_run_map(...)
       error("No current map?")
     end
     local env = dgn_map_meta_wrap(g_dgn_curr_map, dgn)
-    for _, map_chunk_function in pairs(map_chunk_functions) do
+    for _, map_chunk_function in ipairs(map_chunk_functions) do
       if map_chunk_function then
         ret = setfenv(map_chunk_function, env)()
       end
@@ -320,7 +320,7 @@ end
 
 function dgn.fnum_map(map)
   local fnmap = { }
-  for k, v in pairs(map) do
+  for k, v in pairs(map) do -- arbitrary iter order should be ok here
     fnmap[dgn.fnum(k)] = dgn.fnum(v)
   end
   return fnmap
@@ -568,7 +568,7 @@ end
 --   count        - Number of times to repeat the map placement; defaults to 1.
 --                  Note that the map must be marked allow_dup.
 --   die_on_error - If set false, will *not* raise an error if a map cannot
---                  be placed. Defaults to true.
+--                  be placed. Defaults to false.
 -- Returns true if the map(s) were placed successfully, false or nil otherwise.
 --
 -- map, name and tag are mutually exclusive. These parameters may be
@@ -578,18 +578,35 @@ end
 function dgn.place_maps(parameters)
   local n_times = parameters.count or 1
   local n_placed = 0
+
+  local function map_error(name)
+    -- this prefix string is used in map_def::run_hook
+    local e = "Failed to place map '" .. name .. "'"
+    local last_error = dgn.last_builder_error()
+    if last_error and #last_error > 0 then
+      e = e .. "; last builder error: " .. last_error
+    end
+    error(e)
+  end
+
   for i = 1, n_times do
     local map = dgn.find_map(parameters)
     if map then
-      local map_placed_ok = dgn.place_map(map)
+      local map_placed_ok = dgn.place_map(map, true, false, nil, nil,
+                                          parameters.tag)
       if map_placed_ok then
         n_placed = n_placed + 1
       else
         if not parameters.die_on_error then
           return nil
         end
-        error("Failed to place map '" .. dgn.name(map) .. "'")
+        map_error(dgn.name(map))
       end
+    else
+      if not parameters.die_on_error then
+        return nil
+      end
+      error("Could not find map to place")
     end
   end
   return n_placed
@@ -620,37 +637,144 @@ function portal_next(e, next)
   end
 end
 
--- Turn persistant data into a human readable string.
+-- Turn persistent data into a human readable string.
 function persist_to_string()
     return table_to_string(dgn.persist)
 end
 
 -- List of useful scrolls, with some reasonable weights.
--- Attention: total weight is 1000 = 3*120 + 6*70 + 4*40 + 2*15 + 30
 -- When changing the list or the weights, please keep the total weight at 1000.
 dgn.good_scrolls = [[
-    w:80 scroll of identify / scroll of identify q:2 w:30 /
-                               scroll of identify q:3 w:10 /
-    w:80  scroll of teleportation / scroll of teleportation q:2 w:30 /
-                                  / scroll of teleportation q:3 w:10 /
-    w:80  scroll of fog / scroll of fog q:2 w:30 / scroll of fog q:3 w:10 /
-    w:80  scroll of remove curse / scroll of remove curse q:2 w:30 /
-    w:50  scroll of enchant weapon I / scroll of enchant weapon I q:2 w:20 /
-    w:50  scroll of enchant weapon II / scroll of enchant weapon II q:2 w:20 /
-    w:50  scroll of blinking / scroll of blinking q:2 w:20 /
-    w:50  scroll of enchant armour / scroll of enchant armour q:2 w:20 /
-    w:50  scroll of recharging / scroll of recharging q:2 w:20 /
-    w:30  scroll of magic mapping / scroll of magic mapping q:2 w:10 /
-    w:30  scroll of holy word / scroll of holy word q:2 w:10 /
-    w:30  scroll of enchant weapon III / scroll of enchant weapon III q:2 w:10 /
-    w:10  scroll of acquirement / scroll of acquirement q:2 w:4 /
-                                  scroll of acquirement q:3 w:1/
-    w:10   scroll of silence / scroll of silence q:2 w:5 /
-    w:10   scroll of vorpalise weapon /
-    w:10   scroll of immolation /
-    w:10   scroll of vulnerability
+    w:85  scroll of identify no_pickup /
+    w:38  scroll of identify no_pickup q:2 /
+    w:10  scroll of identify no_pickup q:3 /
+    w:85  scroll of teleportation no_pickup /
+    w:38  scroll of teleportation no_pickup q:2 /
+    w:10  scroll of teleportation no_pickup q:3 /
+    w:85  scroll of fog no_pickup /
+    w:33  scroll of fog no_pickup q:2 /
+    w:85  scroll of remove curse no_pickup /
+    w:38  scroll of remove curse no_pickup q:2 /
+    w:95  scroll of enchant weapon no_pickup /
+    w:38  scroll of enchant weapon no_pickup q:2 /
+    w:54  scroll of blinking no_pickup /
+    w:22  scroll of blinking no_pickup q:2 /
+    w:54  scroll of enchant armour no_pickup /
+    w:22  scroll of enchant armour no_pickup q:2 /
+    w:33  scroll of magic mapping no_pickup /
+    w:11  scroll of magic mapping no_pickup q:2 /
+    w:33  scroll of amnesia no_pickup /
+    w:11  scroll of amnesia no_pickup q:2 /
+    w:33  scroll of holy word no_pickup q:1 /
+    w:11  scroll of holy word no_pickup q:2 /
+    w:22  scroll of silence no_pickup q:1 /
+    w:5   scroll of silence no_pickup q:2 /
+    w:11  scroll of acquirement no_pickup q:1 /
+    w:4   scroll of acquirement no_pickup q:2 /
+    w:1   scroll of acquirement no_pickup q:3 /
+    w:11  scroll of brand weapon no_pickup q:1 /
+    w:11  scroll of torment no_pickup q:1 /
+    w:11  scroll of vulnerability no_pickup
     ]]
 
+-- Some scroll and potions with weights that are used as nice loot. These lists
+-- emphasize tactical consumables and permanent character/equipment
+-- upgrades--things that are relevant the entire game and for most characters.
+-- These lists should be used to supplement the more variable loot types % and
+-- * to help ensure that a loot pile has usable items. Each list should total
+-- 100 weight.
+dgn.loot_scrolls = [[
+    w:15  scroll of teleportation /
+    w:15  scroll of fog /
+    w:15  scroll of fear /
+    w:10  scroll of blinking /
+    w:10  scroll of summoning /
+    w:8   scroll of magic mapping /
+    w:10  scroll of enchant weapon /
+    w:10  scroll of enchant armour /
+    w:5   scroll of brand weapon /
+    w:2   scroll of acquirement
+    ]]
+
+dgn.loot_potions = [[
+    w:15  potion of haste /
+    w:15  potion of heal wounds /
+    w:10  potion of might /
+    w:10  potion of invisibility /
+    w:10  potion of magic /
+    w:10  potion of mutation /
+    w:8   potion of cancellation /
+    w:5   potion of brilliance /
+    w:5   potion of resistance /
+    w:2   potion of experience
+    ]]
+
+-- Some definitions for all types of auxiliary armour.
+dgn.aux_armour = "cloak / scarf / helmet / hat / pair of gloves " ..
+    "/ pair of boots"
+
+-- Scarves not influenced by good_item.
+dgn.good_aux_armour = "cloak good_item / scarf / helmet good_item " ..
+    "/ hat good_item / pair of gloves good_item / pair of boots good_item"
+
+-- Scarves excluded since they can't be randart.
+dgn.randart_aux_armour = "cloak randart / helmet randart / hat randart " ..
+    "/ pair of gloves randart / pair of boots randart"
+
+-- Make an item definition that will randomly choose from combinations of the
+-- given tables of weighted item types and optional egos.
+--
+-- @param items     A table with weapon names as keys and weights as values.
+-- @param egos      An optional table with ego names as keys and weights as
+--                  values.
+-- @param args      An optional string of arguments to use on every item entry.
+--                  Should not have leading or trailing whitespace.
+-- @param separator An optional separator to use between the item entries.
+--                  Defaults to '/', which is appropriate for ITEM statements.
+--                  Use '|' if making item statements for use with MONS.
+-- @returns A string containing the item definition.
+function random_item_def(items, egos, args, separator)
+    args = args ~= nil and " " .. args or ""
+    separator = separator ~= nil and separator or '/'
+    local item_def
+    items_list = util.sorted_weight_table(items)
+
+    for i, item_pair in ipairs(items_list) do
+        iname = item_pair[1]
+        iweight = item_pair[2]
+        -- If we have egos, define an item spec with all item+ego
+        -- combinations, each with weight scaled by item rarity and ego
+        -- rarity.
+        if egos ~= nil then
+            egos_list = util.sorted_weight_table(egos)
+            for j, ego_pair in ipairs(egos_list) do
+                ename = ego_pair[1]
+                eweight = ego_pair[2]
+                if (not iname:find("demon") or ename ~= "holy_wrath")
+                   and (not iname:find("quick blade") or ename ~= "speed") then
+                    def = iname .. args .. " ego:" .. ename .. " w:" ..
+                          math.floor(iweight * eweight)
+                    if item_def == nil then
+                         item_def = def
+                    else
+                         item_def = item_def .. " " .. separator .. " " .. def
+                    end
+                end
+            end
+        -- No egos, so define item spec with all item combinations, each with
+        -- weight scaled by item rarity.
+        else
+            def = iname .. args .. " w:" .. iweight
+            if item_def == nil then
+                 item_def = def
+            else
+                 item_def = item_def .. " " .. separator .. " " .. def
+            end
+        end
+    end
+    return item_def
+end
+
 -- Returns true if point1 is inside radius(X, point2).
 function dgn.point_in_radius(point1, point2, radius)
   return dgn.distance(point1.x, point1.y, point2.x, point2.y) <=
